@@ -9,74 +9,6 @@
 
 using CN = dd::ComplexNumbers;
 
-std::string Simulator::MeasureAll(const bool collapse) {
-    if (std::abs(dd::ComplexNumbers::mag2(root_edge.w) - 1.0L) > epsilon) {
-        if (root_edge.w.approximatelyZero()) {
-            throw std::runtime_error("Numerical instabilities led to a 0-vector! Abort simulation!");
-        }
-        std::cerr << "WARNING in MAll: numerical instability occurred during simulation: |alpha|^2 + |beta|^2 = "
-                  << dd::ComplexNumbers::mag2(root_edge.w) << ", but should be 1!\n";
-    }
-
-    dd::Edge cur = root_edge;
-
-    std::string result(getNumberOfQubits(), '0');
-
-    std::uniform_real_distribution<dd::fp> dist(0.0, 1.0L);
-
-    for (dd::Qubit i = root_edge.p->v; i >= 0; --i) {
-        dd::fp p0  = dd::ComplexNumbers::mag2(cur.p->e.at(0).w);
-        dd::fp p1  = dd::ComplexNumbers::mag2(cur.p->e.at(1).w);
-        dd::fp tmp = p0 + p1;
-
-        if (std::abs(tmp - 1.0L) > epsilon) {
-            throw std::runtime_error("Added probabilities differ from 1 by " + std::to_string(std::abs(tmp - 1.0L)));
-        }
-        p0 /= tmp;
-
-        const dd::fp n = dist(mt);
-        if (n < p0) {
-            //result[cur.p->v] = '0';
-            cur = cur.p->e.at(0);
-        } else {
-            result[cur.p->v] = '1';
-            cur              = cur.p->e.at(1);
-        }
-    }
-
-    if (collapse) {
-        dd->decRef(root_edge);
-
-        dd::Package::vEdge                e = dd::Package::vEdge::one;
-        std::array<dd::Package::vEdge, 2> edges{};
-
-        for (dd::Qubit p = 0; p < getNumberOfQubits(); p++) {
-            if (result[p] == '0') {
-                edges[0] = e;
-                edges[1] = dd::Package::vEdge::zero;
-            } else {
-                edges[0] = dd::Package::vEdge::zero;
-                edges[1] = e;
-            }
-            e = dd->makeDDNode(p, edges, false);
-        }
-        dd->incRef(e);
-        root_edge = e;
-        dd->garbageCollect();
-    }
-
-    return std::string{result.rbegin(), result.rend()};
-}
-
-std::map<std::string, std::size_t> Simulator::MeasureAllNonCollapsing(unsigned int shots) {
-    std::map<std::string, std::size_t> results;
-    for (unsigned int i = 0; i < shots; i++) {
-        const auto m = MeasureAll(false);
-        results[m]++;
-    }
-    return results;
-}
-
 std::map<std::string, std::size_t> Simulator::SampleFromAmplitudeVectorInPlace(std::vector<std::complex<dd::fp>>& amplitudes, unsigned int shots) {
     // in-place prefix-sum calculation of probabilities
     std::inclusive_scan(
@@ -154,132 +86,13 @@ void Simulator::NextPath(std::string& s) {
         s.insert(0, "1");
 }
 
-/***
- * Destructively measures a single qubit and returns the result
- * @param index Index of the qubit to be measured
- * @param assume_probability_normalization Probability normalization makes measurement faster but isn't always given
- * @return '1' or '0' depending on the measured value
- */
-char Simulator::MeasureOneCollapsing(const dd::Qubit index, const bool assume_probability_normalization) {
-    std::map<dd::Package::vNode*, dd::fp> probsMone;
-    std::set<dd::Package::vNode*>         visited_nodes2;
-    std::queue<dd::Package::vNode*>       q;
-
-    probsMone[root_edge.p] = CN::mag2(root_edge.w);
-    visited_nodes2.insert(root_edge.p);
-    q.push(root_edge.p);
-
-    while (q.front()->v != index) {
-        dd::Package::vNode* ptr = q.front();
-        q.pop();
-        double prob = probsMone[ptr];
-
-        if (!ptr->e.at(0).w.approximatelyZero()) {
-            const dd::fp tmp1 = prob * CN::mag2(ptr->e.at(0).w);
-
-            if (visited_nodes2.find(ptr->e.at(0).p) != visited_nodes2.end()) {
-                probsMone[ptr->e.at(0).p] = probsMone[ptr->e.at(0).p] + tmp1;
-            } else {
-                probsMone[ptr->e.at(0).p] = tmp1;
-                visited_nodes2.insert(ptr->e.at(0).p);
-                q.push(ptr->e.at(0).p);
-            }
-        }
-
-        if (!ptr->e.at(1).w.approximatelyZero()) {
-            const dd::fp tmp1 = prob * CN::mag2(ptr->e.at(1).w);
-
-            if (visited_nodes2.find(ptr->e.at(1).p) != visited_nodes2.end()) {
-                probsMone[ptr->e.at(1).p] = probsMone[ptr->e.at(1).p] + tmp1;
-            } else {
-                probsMone[ptr->e.at(1).p] = tmp1;
-                visited_nodes2.insert(ptr->e.at(1).p);
-                q.push(ptr->e.at(1).p);
-            }
-        }
-    }
-
-    dd::fp pzero{0}, pone{0};
-
-    if (assume_probability_normalization) {
-        while (!q.empty()) {
-            dd::Package::vNode* ptr = q.front();
-            q.pop();
-
-            if (!ptr->e.at(0).w.approximatelyZero()) {
-                pzero += probsMone[ptr] * CN::mag2(ptr->e.at(0).w);
-            }
-
-            if (!ptr->e.at(1).w.approximatelyZero()) {
-                pone += probsMone[ptr] * CN::mag2(ptr->e.at(1).w);
-            }
-        }
-    } else {
-        std::unordered_map<dd::Package::vNode*, double> probs;
-        assign_probs(root_edge, probs);
-
-        while (!q.empty()) {
-            dd::Package::vNode* ptr = q.front();
-            q.pop();
-
-            if (!ptr->e.at(0).w.approximatelyZero()) {
-                pzero += probsMone[ptr] * probs[ptr->e.at(0).p] * CN::mag2(ptr->e.at(0).w);
-            }
-
-            if (!ptr->e.at(1).w.approximatelyZero()) {
-                pone += probsMone[ptr] * probs[ptr->e.at(1).p] * CN::mag2(ptr->e.at(1).w);
-            }
-        }
-    }
-
-    if (std::abs(pzero + pone - 1) > epsilon) {
-        throw std::runtime_error("Numerical instability occurred during measurement: |alpha|^2 + |beta|^2 = " + std::to_string(pzero) + " + " + std::to_string(pone) + " = " +
-                                 std::to_string(pzero + pone) + ", but should be 1!");
-    }
-
-    const dd::fp sum = pzero + pone;
-
-    //std::pair<fp, fp> probs = std::make_pair(pzero, pone);
-    dd::GateMatrix measure_m{
-            dd::complex_zero, dd::complex_zero,
-            dd::complex_zero, dd::complex_zero};
-
-    std::uniform_real_distribution<dd::fp> dist(0.0, 1.0L);
-    dd::fp                                 n = dist(mt);
-    dd::fp                                 norm_factor;
-    char                                   result;
-
-    if (n < pzero / sum) {
-        measure_m[0] = dd::complex_one;
-        norm_factor  = pzero;
-        result       = '0';
-    } else {
-        measure_m[3] = dd::complex_one;
-        norm_factor  = pone;
-        result       = '1';
-    }
-
-    dd::Package::mEdge m_gate = dd->makeGateDD(measure_m, getNumberOfQubits(), index);
-
-    dd::Package::vEdge e = dd->multiply(m_gate, root_edge);
-
-    dd::Complex c = dd->cn.getTemporary(std::sqrt(1.0 / norm_factor), 0);
-    CN::mul(c, e.w, c);
-    e.w = dd->cn.lookup(c);
-    dd->incRef(e);
-    dd->decRef(root_edge);
-    root_edge = e;
-
-    return result;
-}
-
-double Simulator::ApproximateByFidelity(double targetFidelity, bool allLevels, bool removeNodes, bool verbose) {
+double Simulator::ApproximateByFidelity(std::unique_ptr<dd::Package>& localDD, dd::Package::vEdge& edge, double targetFidelity, bool allLevels, bool removeNodes, bool verbose) {
     std::queue<dd::Package::vNode*>       q;
     std::map<dd::Package::vNode*, dd::fp> probsMone;
 
-    probsMone[root_edge.p] = CN::mag2(root_edge.w);
+    probsMone[edge.p] = CN::mag2(edge.w);
 
-    q.push(root_edge.p);
+    q.push(edge.p);
 
     while (!q.empty()) {
         dd::Package::vNode* ptr = q.front();
@@ -303,9 +116,9 @@ double Simulator::ApproximateByFidelity(double targetFidelity, bool allLevels, b
         }
     }
 
-    std::vector<int>                                                                                                              nodes(getNumberOfQubits(), 0);
-    std::vector<std::priority_queue<std::pair<double, dd::Package::vNode*>, std::vector<std::pair<double, dd::Package::vNode*>>>> qq(
-            getNumberOfQubits());
+    std::vector<int> nodes(getNumberOfQubits(), 0);
+
+    std::vector<std::priority_queue<std::pair<double, dd::Package::vNode*>, std::vector<std::pair<double, dd::Package::vNode*>>>> qq(getNumberOfQubits());
 
     for (auto& it: probsMone) {
         if (it.first->v < 0) {
@@ -350,21 +163,21 @@ double Simulator::ApproximateByFidelity(double targetFidelity, bool allLevels, b
         dag_edges[it] = dd::Package::vEdge::zero;
     }
 
-    dd::Package::vEdge newEdge = RemoveNodes(root_edge, dag_edges);
-    assert(!std::isnan(dd::CTEntry::val(root_edge.w.r)));
-    assert(!std::isnan(dd::CTEntry::val(root_edge.w.i)));
-    dd::Complex c = dd->cn.getCached(std::sqrt(CN::mag2(newEdge.w)), 0);
+    dd::Package::vEdge newEdge = RemoveNodes(localDD, edge, dag_edges);
+    assert(!std::isnan(dd::CTEntry::val(edge.w.r)));
+    assert(!std::isnan(dd::CTEntry::val(edge.w.i)));
+    dd::Complex c = localDD->cn.getCached(std::sqrt(CN::mag2(newEdge.w)), 0);
     CN::div(c, newEdge.w, c);
-    newEdge.w = dd->cn.lookup(c);
+    newEdge.w = localDD->cn.lookup(c);
 
     dd::fp fidelity = 0;
-    if (root_edge.p->v == newEdge.p->v) {
-        fidelity = dd->fidelity(root_edge, newEdge);
+    if (edge.p->v == newEdge.p->v) {
+        fidelity = localDD->fidelity(edge, newEdge);
     }
 
     if (verbose) {
-        const unsigned size_before = dd->size(root_edge);
-        const unsigned size_after  = dd->size(newEdge);
+        const unsigned size_before = localDD->size(edge);
+        const unsigned size_after  = localDD->size(newEdge);
         std::cout
                 << getName() << ","
                 << +getNumberOfQubits() << "," // unary plus for int promotion
@@ -380,22 +193,22 @@ double Simulator::ApproximateByFidelity(double targetFidelity, bool allLevels, b
     }
 
     if (removeNodes) {
-        dd->decRef(root_edge);
-        dd->incRef(newEdge);
-        root_edge = newEdge;
+        localDD->decRef(edge);
+        localDD->incRef(newEdge);
+        edge = newEdge;
     }
     return fidelity;
 }
 
-double Simulator::ApproximateBySampling(unsigned int nSamples, unsigned int threshold, bool removeNodes, bool verbose) {
+double Simulator::ApproximateBySampling(std::unique_ptr<dd::Package>& localDD, dd::Package::vEdge& edge, std::size_t nSamples, std::size_t threshold, bool removeNodes, bool verbose) {
     assert(nSamples > threshold);
     std::map<dd::Package::vNode*, unsigned int> visited_nodes;
     std::uniform_real_distribution<dd::fp>      dist(0.0, 1.0L);
 
     for (unsigned int j = 0; j < nSamples; j++) {
-        dd::Edge cur = root_edge;
+        dd::Edge cur = edge;
 
-        for (dd::Qubit i = root_edge.p->v; i >= 0; --i) {
+        for (dd::Qubit i = edge.p->v; i >= 0; --i) {
             visited_nodes[cur.p]++;
 
             dd::fp p0 = CN::mag2(cur.p->e.at(0).w);
@@ -410,9 +223,9 @@ double Simulator::ApproximateBySampling(unsigned int nSamples, unsigned int thre
     }
 
     std::set<dd::Package::vNode*> visited_nodes2;
-    visited_nodes2.insert(root_edge.p);
+    visited_nodes2.insert(edge.p);
     std::queue<dd::Package::vNode*> q;
-    q.push(root_edge.p);
+    q.push(edge.p);
 
     while (!q.empty()) {
         dd::Package::vNode* ptr = q.front();
@@ -440,19 +253,19 @@ double Simulator::ApproximateBySampling(unsigned int nSamples, unsigned int thre
         dag_edges[it] = dd::Package::vEdge::zero;
     }
 
-    dd::Package::vEdge newEdge = RemoveNodes(root_edge, dag_edges);
-    dd::Complex        c       = dd->cn.getCached(std::sqrt(CN::mag2(newEdge.w)), 0);
+    dd::Package::vEdge newEdge = RemoveNodes(localDD, edge, dag_edges);
+    dd::Complex        c       = localDD->cn.getCached(std::sqrt(CN::mag2(newEdge.w)), 0);
     CN::div(c, newEdge.w, c);
-    newEdge.w = dd->cn.lookup(c);
+    newEdge.w = localDD->cn.lookup(c);
 
     dd::fp fidelity = 0;
-    if (root_edge.p->v == newEdge.p->v) {
-        fidelity = dd->fidelity(root_edge, newEdge);
+    if (edge.p->v == newEdge.p->v) {
+        fidelity = localDD->fidelity(edge, newEdge);
     }
 
     if (verbose) {
-        const unsigned size_after  = dd->size(newEdge);
-        const unsigned size_before = dd->size(root_edge);
+        const unsigned size_after  = localDD->size(newEdge);
+        const unsigned size_before = localDD->size(edge);
         std::cout
                 << getName() << ","
                 << +getNumberOfQubits() << "," // unary plus for int promotion
@@ -468,17 +281,16 @@ double Simulator::ApproximateBySampling(unsigned int nSamples, unsigned int thre
     }
 
     if (removeNodes) {
-        dd->decRef(root_edge);
-        dd->incRef(newEdge);
-        root_edge = newEdge;
-        dd->garbageCollect();
+        localDD->decRef(edge);
+        localDD->incRef(newEdge);
+        edge = newEdge;
+        localDD->garbageCollect();
     }
 
     return fidelity;
 }
 
-dd::Package::vEdge
-Simulator::RemoveNodes(dd::Package::vEdge e, std::map<dd::Package::vNode*, dd::Package::vEdge>& dag_edges) {
+dd::Package::vEdge Simulator::RemoveNodes(std::unique_ptr<dd::Package>& localDD, dd::Package::vEdge e, std::map<dd::Package::vNode*, dd::Package::vEdge>& dag_edges) {
     if (e.isTerminal()) {
         return e;
     }
@@ -489,21 +301,21 @@ Simulator::RemoveNodes(dd::Package::vEdge e, std::map<dd::Package::vNode*, dd::P
         if (r.w.approximatelyZero()) {
             return dd::Package::vEdge::zero;
         }
-        dd::Complex c = dd->cn.getTemporary();
+        dd::Complex c = localDD->cn.getTemporary();
         dd::ComplexNumbers::mul(c, e.w, r.w);
-        r.w = dd->cn.lookup(c);
+        r.w = localDD->cn.lookup(c);
         return r;
     }
 
     std::array<dd::Package::vEdge, dd::RADIX> edges{
-            RemoveNodes(e.p->e.at(0), dag_edges),
-            RemoveNodes(e.p->e.at(1), dag_edges)};
+            RemoveNodes(localDD, e.p->e.at(0), dag_edges),
+            RemoveNodes(localDD, e.p->e.at(1), dag_edges)};
 
-    dd::Package::vEdge r = dd->makeDDNode(e.p->v, edges, false);
+    dd::Package::vEdge r = localDD->makeDDNode(e.p->v, edges, false);
     dag_edges[e.p]       = r;
-    dd::Complex c        = dd->cn.getCached();
+    dd::Complex c        = localDD->cn.getCached();
     CN::mul(c, e.w, r.w);
-    r.w = dd->cn.lookup(c);
+    r.w = localDD->cn.lookup(c);
     return r;
 }
 
@@ -542,20 +354,3 @@ std::pair<dd::ComplexValue, std::string> Simulator::getPathOfLeastResistance() c
     return {{dd::CTEntry::val(path_value.r), dd::CTEntry::val(path_value.i)},
             std::string{result.rbegin(), result.rend()}};
 }
-
-double Simulator::assign_probs(dd::Package::vEdge edge, std::unordered_map<dd::Package::vNode*, double>& probs) {
-    auto it = probs.find(edge.p);
-    if (it != probs.end()) {
-        return CN::mag2(edge.w) * it->second;
-    }
-    double sum;
-    if (edge.isTerminal()) {
-        sum = 1.0;
-    } else {
-        sum = assign_probs(edge.p->e.at(0), probs) + assign_probs(edge.p->e.at(1), probs);
-    }
-
-    probs.insert(std::pair<dd::Package::vNode*, double>(edge.p, sum));
-
-    return CN::mag2(edge.w) * sum;
-}
